(define-module lmn.peg
  (use util.stream)
  (use util.queue)
  (export <parse-error>
          parse-stream parse-port parse-string
          $ret $eof $c $any $s
          $<< $or $look $unexpect $skip $g
          $delay $let $let1 $do* $do
          $none-of
          $cons $xcons $list $cons* $nth $seq $seq0
          $optional $>>
          $many $many1 $rep $many-till $many1-till
          $between
          $followed-by $followed-by1 $led-by $led-by1
          $sep-by $sep-by1 $alternate1 $alternate
          $chain-left $chain-right $fold-left $fold-right))

(select-module lmn.peg)

;; For Emacsen (evaluate with C-x C-e) :
;; (outline-minor-mode 1)
;; (setq-local outline-regexp "^[\s\t]*;;[\s]+\\+[+-]*\s")
;; (setq-local outline-level (lambda () (- (outline-level) 4)))

;; a parser takes two arguments
;;
;; - target stream
;; - string representation of the last successful object or (*)
;;
;; and return
;;
;; - #t
;; - a pair of
;;   - the parsed object
;;   - string representation of the parsed object (*)
;; - unread stream
;;
;; on success, or
;;
;; - #f
;; - a dotted list of
;;   - string representation of the expected item, or a list of it (**)
;;   - string representation the last successful object (*)
;;   - string representation of what actually got, or a dotted list recursively
;; - unread stream
;;
;; on failure.
;;
;; (*) ... can be #f iff nothing ever succeeded
;; (**) ... can be omtted by returning #f

;; 下の例でエラーメッセージが変
;;
;; (use srfi-13) ;; string>
;;
;; (define $comment
;;   ($or ($seq ($or ($c #\%) ($s "//"))
;;              ($many ($none-of #[\n]))
;;              ($or ($c #\newline) $eof))
;;        ($seq ($s "/*")
;;              ($many-till ($any) ($s "*/")))))
;;
;; (define ($token parser)
;;   ($seq0 parser ($many ($or $comment ($any #[\x09-\x0d ])))))
;;
;; (define (make-expr-parser term ops)
;;   ;;  Make a parser that parses a pre/in-fix expression. OPS must be a
;;   ;; list of operators of the form (PRIORITY TYPE OP OP ...) where
;;   ;; PRIORITY is an integer and TYPE is one of 'left 'right or
;;   ;; 'prefix. OP is a list of the form (STR FN) for infix operators or
;;   ;; (STR FN . ARITY) for prefix operators, where ARITY is a natural
;;   ;; number or nil. If ARITY is nil, the operator is counted as an
;;   ;; unray operator.
;;   (letrec ([expr
;;             (let loop ([ops (sort ops (^(l1 l2) (< (car l1) (car l2))))])
;;               (if (null? ops)
;;                   ($or ($between ($token ($c #\())
;;                                  ($g ($delay expr) "expression")
;;                                  ($token ($c #\))))
;;                        term)
;;                   (let ([$subexp ($g (loop (cdr ops)) "term")]
;;                         [$op (apply $or (map (^o ($seq ($s (car o)) ($ret (cdr o))))
;;                                              (sort (cddar ops) (^(o1 o2) (string> (car o1) (car o2))))))])
;;                     (case (cadar ops)
;;                       [(prefix)
;;                        ($or ($g ($do* op <- ($token $op)
;;                                       terms <- ($rep $subexp (if (integer? (cdr op)) (cdr op) 1))
;;                                       ($ret (apply (car op) terms)))
;;                                 "prefix operation" #t)
;;                             $subexp)]
;;                       [(left)
;;                        ($chain-left $subexp ($token ($>> $op car)))]
;;                       [(right)
;;                        ($chain-right $subexp ($token ($>> $op car)))]))))])
;;     expr))
;;
;; (parse-string
;;  (letrec ([term ($or ($token ($do* x <- ($many1 ($any #[0-9]))
;;                                    (let1 val (string->number (list->string x))
;;                                      ($ret val #`"integer literal ,val"))))
;;                      ($g ($between ($token ($c #\[))
;;                                    ($sep-by ($delay expr) ($token ($c #\,)))
;;                                    ($token ($c #\])))
;;                          "list"))]
;;           [expr (make-expr-parser
;;                  term
;;                  `((0 left
;;                       ("times" ,*))
;;                    (1 prefix
;;                       ("twice" ,(^x (* x 2)))
;;                       ("multiply" ,* . 2))
;;                    (100 left
;;                         ("*" ,*)
;;                         ("/" ,/))
;;                    (200 right
;;                         ("^" ,expt))
;;                    (50 left
;;                        ("-" ,-)
;;                        ("+" ,+)
;;                        ("++" ,append!))
;;                    (1000 prefix
;;                          ("-" ,-)
;;                          ("+/" ,(^x (apply + x))))
;;                    ))])
;;    expr)
;;  "twice (-)")

;; "while parsing ~" ほしいかな？やっぱり

;; ---- parser drivers

(define-condition-type <parse-error> <error> #f)

(define (parse-error->string value)
  (let* ([expecteds->string
          (^(expecteds)
            (cond [(not expecteds) #f]
                  [(pair? expecteds) #`"one of { ,(string-join expecteds \", \") }"]
                  [else expecteds]))]
         [after
          (if-let1 l (cadr value) #`" after ,l" "")]
         [actual
          (let1 actual (caddr value)
            (if (string? actual)
                actual
                (let loop ([actual actual])
                  (cond [(caddr actual) pair? => loop]
                        [else (let ([after (if-let1 l (cadr actual) #`" after ,l" "")]
                                    [expected (if-let1 s (expecteds->string (car actual))
                                                #`" (where ,s is expected)" "")])
                                #`",(caddr actual),after,expected")]))))])
    (if-let1 s (expecteds->string (car value))
      #`"expected ,s,after but got ,actual"
      #`"unexpected ,actual,after")))

(define (test-parser parser str)
  ;; A test driver for parsers.
  (receive (status value rest) (parser (string->stream str) #f)
    (cond [status
           (print "PARSE SUCCESS:")
           (print "     object: " (car value))
           (print "description: " (cdr value))
           (print "     unread: " (stream->string rest))]
          [else
           (print "PARSE FAILURE:")
           (print "expected: " (car value))
           (print "   after: " (cadr value))
           (print "  actual: " (caddr value))
           (print "  unread: " (stream->string rest))])))

;; Parser<T> -> Stream<Char> -> T
(define (parse-stream parser stream)
  ;; Apply PARSER to STREAM and return the result.
  (receive (status value _) (parser stream #f)
    (if status
        (car value)
        (error <parse-error> (parse-error->string value)))))

;; Parser<T> [-> Port] -> T
(define (parse-port parser :optional [port (current-input-port)])
  ;; Apply PARSER to PORT. When PORT is omitted, apply to the stdin.
  (parse-stream parser (port->stream)))

;; Parser<T> -> String -> T
(define (parse-string parser string)
  ;; Apply PARSER to STRING.
  (parse-stream parser (string->stream string)))

;; ---- 5 core parsers

;; T -> Parser<T>
(define (($ret obj :optional [name #f]) stream last)
  ;; A parser that just returns OBJ without consuming any
  ;; characters. You can optionally add a description for the returned
  ;; value to improve error messages.
  (values #t (cons obj (or name last)) stream))

;; (test-parser ($ret "hoge") "fuga")
;; (test-parser ($ret "hoge" "HOGE") "fuga")

;; Parser<()>
(define ($eof stream last)
  ;; A parser that succeeds only at the end of input.
  (if (stream-pair? stream)
      (values #f (list "EOF" last #`"',(stream-car stream)'") stream)
      (values #t (cons (eof-object) "EOF") stream)))

;; (test-parser $eof "")
;; (test-parser $eof "hoge")

;; Char -> Parser<Char>
(define (($c ch) stream last)
  ;; A parser that parses a character which equals to CH.
  (cond [(stream-null? stream)
         (values #f (list #`"',ch'" last "EOF") stream)]
        [(char=? (stream-car stream) ch)
         (values #t (cons ch #`"',ch'") (stream-cdr stream))]
        [else
         (values #f (list #`"',ch'" last #`"',(stream-car stream)'") stream)]))

;; (test-parser ($c #\a) "abc")
;; (test-parser ($c #\a) "bcd")
;; (test-parser ($c #\a) "")

;; CharSet -> Parser<Char>
(define (($any :optional [set #[^]]) stream last)
  ;; A parser that parses a character which belongs SET.
  (cond [(stream-null? stream)
         (values #f (list #`"any ,set" last "EOF") stream)]
        [(stream-car stream) (pa$ char-set-contains? set) =>
         (^c (values #t (cons c #`"',c'") (stream-cdr stream)))]
        [else
         (values #f (list #`"any ,set" last #`"',(stream-car stream)'") stream)]))

;; (test-parser ($any #[a-z]) "abc")
;; (test-parser ($any #[0-9]) "abc")
;; (test-parser ($any #[a-z]) "")
;; (test-parser ($any) "abc")

;; String -> Parser<String>
(define (($s str) stream last)
  ;; A parser that parses a string which equals to STR.
  (let loop ([s stream] [expected (string->list str)] [actual ()])
    (cond [(null? expected)
           (values #t (cons str #`"',str'") s)]
          [(stream-null? s)
           (values
            #f (list #`"',str'" last #`",(list->string (reverse! actual))[EOF]") stream)]
          [else
           (let1 ch (stream-car s)
             (if (char=? (car expected) ch)
                 (loop (stream-cdr s) (cdr expected) (cons ch actual))
                 (values
                  #f (list #`",str" last #`",(list->string (reverse! actual))',ch'") stream)))])))

;; (test-parser ($s "abc") "abcdefg")
;; (test-parser ($s "abc") "abxdefg")
;; (test-parser ($s "abc") "ab")

;; ---- 6 core combinators

;; (A -> B -> ... Y -> Z) -> Parser<A> -> Parser<B> -> ... -> Parser<Y> -> Parser<Z>
(define (($<< fn :rest parsers) stream last)
  ;; Use PARSERS in sequence and apply results to FN, to make the
  ;; final result.
  (let loop ([parsers parsers] [stream stream] [last last] [results ()])
    (if (null? parsers)
        (values #t (cons (apply fn (reverse! results)) last) stream)
        (receive (status value rest) ((car parsers) stream last)
          (if status
              (loop (cdr parsers) rest (cdr value) (cons (car value) results))
              (values #f value rest))))))

;; (test-parser ($<< list ($s "hoge") ($s "fuga")) "hogefuga")
;; (test-parser ($<< list ($s "hoge") ($s "fuga")) "hogehoge")

;; Parser<T> -> Parser<T> -> ... -> Parser<T> -> Parser<T>
(define (($or parser :rest parsers) stream last)
  ;; Use the first PARSER to parse an object. If it succeeds or
  ;; consumes any characters, just return the result or
  ;; fail. Otherwise, that is, if it fails without consuming any
  ;; characters, try another PARSER to parse an object if there is
  ;; one, or just fail.
  (let loop ([parsers (cons parser parsers)] [expecteds ()] [actuals ()])
    (if (null? parsers)
        ;; *FIXME* algorithm for searching the deppest error should be changed
        (let1 actual (let loop ([v (car actuals)] [q (list->queue (cdr actuals))])
                       (cond [(queue-empty? q)
                              v]
                             [(string? v)
                              (loop (dequeue! q) q)]
                             [else
                              (enqueue! q (caddr v))
                              (loop (dequeue! q) q)]))
          (values #f (list expecteds last actual) stream))
        (receive (status value rest) ((car parsers) stream last)
          (if (or status (not (eq? stream rest)))
              (values status value rest)
              (loop (cdr parsers)
                    ((if (pair? (car value)) append! cons) (car value) expecteds)
                    (cons (caddr value) actuals)))))))

;; (test-parser ($<< list ($s ":") ($or ($s "hoge") ($s "fuga"))) ":hogehoge")
;; (test-parser ($<< list ($s ":") ($or ($s "hoge") ($s "fuga"))) ":fugafuga")
;; (test-parser ($<< list ($s ":") ($or ($s "hoge") ($s "fuga"))) ":piyopiyo")

;; Parser<T> -> Parser<T>
(define (($look parser) stream last)
  ;; Use PARSER to parse an object. On success, pretend as if no
  ;; characters were consumed.
  (receive (status value rest) (parser stream last)
    (if status
        (values #t (cons (car value) last) stream)
        (values #f value rest))))

;; (test-parser ($<< list ($s "hoge") ($look ($<< list ($s "a") ($s "b")))) "hogeab")
;; (test-parser ($<< list ($s "hoge") ($look ($<< list ($s "a") ($s "b")))) "hogeax")

(define (($unexpect parser :optional [name #f]) stream last)
  ;; Negative version of "$look". Succeed and pretend as if no
  ;; characters were consumed, iff PARSER fails.
  (receive (status value rest) (parser stream last)
    (if status
        (values #f (list #f last (or name (cdr value))) rest)
        (values #t (cons (undefined) last) stream))))

;; (test-parser ($<< list ($s ":") ($unexpect ($<< list ($s "a") ($s "b")) "ab")) ":ax")
;; (test-parser ($<< list ($s ":") ($unexpect ($<< list ($s "a") ($s "b")) "ab")) ":ab")

(define (($skip parser) stream last)
  ;; Parse an object with PARSER without affecting error messages.
  (receive (status value rest) (parser stream last)
    (if status
        (values #t (cons (car value) last) rest)
        (values #f value rest))))

;; (test-parser ($<< list ($skip ($s "hoge")) ($s "fuga")) "hogefuga")
;; (test-parser ($<< list ($skip ($s "hoge")) ($s "fuga")) "hogehoge")

;; String -> Parser<T> -> Parser<T>
(define (($g parser :optional [name #f] [atomic #f]) stream last)
  ;; Group a (complex) parser PARSER. You can give a name to the group
  ;; to improve error messages, and/or make the group atomic to allow
  ;; backtracking.
  (receive (status value rest) (parser stream last)
    (cond [status ;; PARSER succeeded
           (values #t value rest)]
          [(eq? stream rest) ;; PARSER failed without consuming any characters
           (values
            #f (list (or name (car value)) last (caddr value)) rest)]
          [atomic ;; PARSER failed after consuming some chars, but ATOMIC = #t
           (values
            #f (list name last (if (pair? (caddr value)) (caddr value) value)) stream)]
          [else
           (values #f value rest)])))

;; non-atomic
;; (test-parser ($g ($<< list ($s ":") ($g ($<< list ($s "a") ($s "b")) "ab")) ":") ":ab")
;; (test-parser ($g ($<< list ($s ":") ($g ($<< list ($s "a") ($s "b")) "ab")) ":") ":ax")
;; (test-parser ($g ($<< list ($s ":") ($g ($<< list ($s "a") ($s "b")) "ab")) ":") ":x")
;; (test-parser ($g ($<< list ($s ":") ($g ($<< list ($s "a") ($s "b")) "ab")) ":") "x")

;; atomic
;; (test-parser ($g ($<< list ($s ":") ($g ($<< list ($s "a") ($s "b")) "ab" #t)) ":") ":ab")
;; (test-parser ($g ($<< list ($s ":") ($g ($<< list ($s "a") ($s "b")) "ab" #t)) ":") ":ax")
;; (test-parser ($g ($<< list ($s ":") ($g ($<< list ($s "a") ($s "b")) "ab" #t)) ":") ":x")
;; (test-parser ($g ($<< list ($s ":") ($g ($<< list ($s "a") ($s "b")) "ab" #t)) ":") "x")

;; controling backtrack
;; (test-parser ($or ($g ($<< list ($c #\1) ($c #\a)) "1a"   ) ($c #\1)) "1")
;; (test-parser ($or ($g ($<< list ($c #\1) ($c #\a)) "1a" #t) ($c #\1)) "1")
;; (test-parser ($or ($g ($<< list ($c #\1) ($c #\a)) "1a" #t) ($c #\1)) "2")

;; ---- utility macros

(define-macro ($delay parser)
  ;; Make a thunk equivalent to PARSER. This may be useful when
  ;; writing mutually recursive parsers.
  `(^(s l) (,parser s l)))

(define-macro ($let vars :rest body)
  ;; example:
  ;; ($let ([x foo] [y bar]) (cons x y))
  ;; = ($<< (^(x y) (cons x y)) foo bar)
  `($<< (^,(map car vars) ,@body) ,@(map cadr vars)))

(define-macro ($let1 var parser :rest body)
  ;; example:
  ;; ($let1 x foo (list->string x))
  ;; = ($<< (^x (list->string x)) foo)
  `($<< (^(,var) ,@body) ,parser))

(define-macro ($do* :rest body)
  ;; "Do Notation" for parsers. See also "$do".
  ;; example: ($do x <- ($any #[a-z])
  ;;               y <- ($c (char-upcase x))
  ;;              ($ret (string x y)))
  (let loop ([body body] [vars ()] [parsers ()])
    (cond [(null? body)
           `(^(s l) (let ,(map (cut list <> (undefined)) (delete-duplicates! vars eq?))
                      (($<< ,(^(:rest xs) (last xs)) ,@(reverse! parsers)) s l)))]
          [(eq? (list-ref body 1 #f) '<-)
           (let1 s (gensym)
             (loop (cdddr body)
                   (cons (car body) vars)
                   (cons `($<< (^(,s) (set! ,(car body) ,s) ,s) ($delay ,(caddr body)))
                         parsers)))]
          [else
           (loop (cdr body) vars (cons `($delay ,(car body)) parsers))])))

(define-macro ($do :rest body)
  ;; Like "$do*" but bound variables are only visible in the last
  ;; parser and the last parser must be ($ret <>).
  ;; example:
  ;; ($do x <- foo y <- bar ($ret (cons x y)))
  ;; = ($let ([x foo] [y bar]) (cons x y))
  (let loop ([body body] [vars ()] [parsers ()])
    (cond [(null? body)
           (error "$do with no parsers")]
          [(null? (cdr body))
           (if (eq? (caar body) '$ret)
               `($<< (^,(reverse! vars) ,(cadar body)) ,@(reverse! parsers))
               (error "$do must be terminated with $ret. try $do*."))]
          [(eq? (cadr body) '<-)
           (loop (cdddr body) (cons (car body) vars) (cons (caddr body) parsers))]
          [else
           (loop (cdr body) (cons '_ vars) (cons (car body) parsers))])))

;; ---- library parsers/combinators

;; inverse of "$any"
(define ($none-of charset) ($any (char-set-complement charset)))

;; pair
(define ($cons a d)  ($<< cons a d))
(define ($xcons d a) ($<< (^(d a) (cons a d)) d a))

;; list
(define ($list :rest parsers)  (apply $<< list parsers))
(define ($cons* :rest parsers) (apply $<< cons* parsers))
(define ($nth n :rest parsers) (apply $<< (^(:rest xs) (list-ref xs n)) parsers))
(define ($seq :rest parsers)   (apply $<< (^(:rest xs) (last xs)) parsers))
(define ($seq0 :rest parsers)  (apply $<< (^(x :rest _) x) parsers))

;; Parser<T> -> T -> Parser<T>
(define ($optional parser :optional default) ($or parser ($ret default)))

;; Parser<A> -> (A -> B) -> (B -> C) -> ... -> (Y -> Z) -> Parser<Z>
(define ($>> parser :rest functions) ($<< (apply compose (reverse! functions)) parser))

;; AA... => (A A ...)
(define ($many parser)
  (letrec ([self ($or ($cons parser ($delay self)) ($ret ()))]) self))
(define ($many1 parser) ($cons parser ($many parser)))
(define ($rep parser n) (apply $list (make-list n parser)))

;; AA...AB => (A A ... A)
(define ($many-till term delim)  ($seq0 ($many ($seq ($unexpect delim) term)) delim))
(define ($many1-till term delim) ($cons term ($many-till term delim)))

;; ABC => B
(define ($between left term right) ($<< (^(_ x _) x) left term right))

;; ABAB...AB => (A A ... A)
(define ($followed-by term delim)  ($many ($seq0 term delim)))
(define ($followed-by1 term delim) ($many1 ($seq0 term delim)))

;; BABA...BA => (A A ... A)
(define ($led-by term delim)  ($many ($seq delim term)))
(define ($led-by1 term delim) ($many1 ($seq delim term)))

;; ABAB...BA => (A A ... A)
(define ($sep-by1 term delim) ($cons term ($led-by term delim)))
(define ($sep-by term delim)  ($optional ($sep-by1 term delim) ()))

;; ABAB...BA => (A B A B ... B A)
(define ($alternate1 term delim)
  ($cons term ($<< (pa$ apply append) ($many ($list delim term)))))
(define ($alternate term delim)  ($optional ($alternate1 term delim) ()))

;; binary operators :: Parser<A> -> Parser<A->A->A> -> Parser<A>
(define ($chain-left term op)
  ($<< (^(i l) (fold-left (^(x y) ((car y) x (cdr y))) i l))
       term ($many ($cons op term))))
(define ($chain-right term op)
  ($<< (^(l i) (fold-right (^(x y) ((car x) (cdr x) y)) i l))
       ($many ($g ($xcons term op) #f #t)) term))

;; folding parser :: (B[A]->A[B]->B) -> B -> Parser<A> -> Parser<B>
(define ($fold-left fn init parser)  ($<< (pa$ fold-left fn init) ($many parser)))
(define ($fold-right fn init parser) ($<< (pa$ fold-right fn init) ($many parser)))
